r"""This module defines the base model."""

from __future__ import annotations

__all__ = ["attach_module_to_engine", "is_model_config", "setup_model", "setup_and_attach_model"]

import logging

from objectory.utils import is_object_config
from torch.nn import Module

from gravitorch.engines.base import BaseEngine
from gravitorch.models.base import BaseModel
from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def attach_module_to_engine(module: Module, engine: BaseEngine) -> None:
    r"""Attaches a module to the engine if the module has the ``attach``
    method.

    This function does nothing if the module does not have a
    ``attach`` method.

    Args:
    ----
        module (``torch.nn.Module``): Specifies the module to attach
            to the engine.
        engine (``BaseEngine``): Specifies the engine.
    """
    if hasattr(module, "attach"):
        module.attach(engine)


def is_model_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseModel``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``BaseModel`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.models import is_model_config
        >>> is_model_config({"_target_": "gravitorch.models.VanillaModel"})
        True
    """
    return is_object_config(config, BaseModel)


def setup_model(model: Module | dict) -> Module:
    r"""Sets up the model.

    The model is instantiated from its configuration by using the
    ``BaseModel`` factory function.

    Args:
    ----
        model (``torch.nn.Module`` or dict): Specifies the model or
            its configuration.

    Returns:
    -------
        ``torch.nn.Module``: The (instantiated) model.
    """
    if isinstance(model, dict):
        logger.info(f"Initializing a model from its configuration... {str_target_object(model)}")
        model = BaseModel.factory(**model)
    return model


def setup_and_attach_model(engine: BaseEngine, model: Module | dict) -> Module:
    r"""Sets up and attaches the model to the engine.

    The model is attached to the engine by using the ``attach`` method.
     If the model does not have a ``attach`` method, the ``attach``
     step is skipped.

    Note that if you call this function ``N`` times with the same
    model, the model will be attached ``N`` times to the engine.

    Args:
    ----
        engine (``BaseEngine``): Specifies the engine.
        model (``torch.nn.Module`` or dict): Specifies the model or
            its configuration.

    Returns:
    -------
        ``torch.nn.Module``: The (instantiated) model.
    """
    model = setup_model(model)
    if hasattr(model, "attach"):
        logger.info("Adding a model to the engine state...")
        model.attach(engine=engine)
    return model



from __future__ import annotations

__all__ = ["BaseDataFlow"]

from abc import ABC, abstractmethod
from collections.abc import Iterator
from types import TracebackType
from typing import Generic, TypeVar

T = TypeVar("T")


class BaseDataFlow(Generic[T], ABC):
    r"""Base class to implement a dataflow.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.experimental.dataflow import IterableDataFlow
        >>> with IterableDataFlow([1, 2, 3, 4, 5]) as dataflow:
        ...     for batch in dataflow:
        ...         print(batch)  # do something
        ...
    """

    def __enter__(self) -> BaseDataFlow:
        self.launch()
        return self

    def __exit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        self.shutdown()

    @abstractmethod
    def __iter__(self) -> Iterator[T]:
        r"""Returns an iterator on the data."""

    @abstractmethod
    def launch(self) -> None:
        r"""Launch the data flow.

        Example usage:

        .. code-block:: pycon

            >>> dataflow: BaseDataFlow = ...
            >>> dataflow.launch()
        """

    @abstractmethod
    def shutdown(self) -> None:
        r"""Shutdown the data flow.

        Example usage:

        .. code-block:: pycon

            >>> dataflow: BaseDataFlow = ...
            >>> dataflow.launch()
            >>> # do anything
            >>> dataflow.shutdown()
        """



from __future__ import annotations

__all__ = ["ImageNetDataSource", "create_train_eval_datasets_v1"]

import logging
from pathlib import Path

from torch.utils.data import Dataset

from gravitorch import constants as ct
from gravitorch.creators.dataloader.base import BaseDataLoaderCreator
from gravitorch.data.datasets.image_folder import ImageFolderDataset
from gravitorch.datasources.dataset import DatasetDataSource
from gravitorch.utils.imports import check_torchvision, is_torchvision_available

if is_torchvision_available():
    from torchvision import transforms
else:
    transforms = None  # pragma: no cover

logger = logging.getLogger(__name__)

IMAGENET_DEFAULT_MEAN = (0.485, 0.456, 0.406)
IMAGENET_DEFAULT_STD = (0.229, 0.224, 0.225)


class ImageNetDataSource(DatasetDataSource):
    r"""Implements a datasource for the ImageNet dataset."""

    @classmethod
    def create_imagenet_v1(
        cls,
        train_path: Path | str | None,
        eval_path: Path | str | None,
        dataloader_creators: dict[str, BaseDataLoaderCreator | dict | None],
        input_size: int = 224,
    ) -> ImageNetDataSource:
        r"""Creates a datasource for the ImageNet dataset the training
        and evaluation datasets with their original data augmentations
        (V1).

        This function initializes the original data augmentations,
        sometimes called V1
        (https://pytorch.org/blog/introducing-torchvision-new-multi-weight-support-api/).
        Note that more aggressive data augmentation can be used to
        improve the performances. See the following article for more
        information:

            Wightman R., Touvron H., Jégou H.
            ResNet strikes back: An improved training procedure in timm.
            http://arxiv.org/pdf/2110.00476

        Args:
        ----
            train_path (``pathlib.Path`` or str or ``None``):
                Specifies the path to the training dataset.
                If ``None``, the training dataset is set to ``None``.
            eval_path (``pathlib.Path`` or str or ``None``):
                Specifies the path to the evaluation dataset.
                If ``None``, the evaluation dataset is set to ``None``.
            dataloader_creators: (dict): Specifies the data loader
                creators to initialize. Each key indicates a data
                loader creator name. For example if you want to create
                a data loader for ``'train'`` ID, the dictionary has to
                have a key ``'train'``. The value can be a
                ``BaseDataLoaderCreator`` object, or its configuration,
                or ``None``. ``None`` means a default data loader will
                be created. Each data loader creator takes a
                ``Dataset`` object as input, so you need to specify
                the associated dataset path.
            input_size (int, optional): Specifies the input size for
                the image. Default: ``224``

        Returns:
        -------
            ``ImageNetDataSource``: A datasource for the ImageNet datasets.
        """
        train_dataset, eval_dataset = create_train_eval_datasets_v1(
            train_path=train_path,
            eval_path=eval_path,
            input_size=input_size,
        )
        datasets = {}
        if train_dataset:
            datasets[ct.TRAIN] = train_dataset
        if eval_dataset:
            datasets[ct.EVAL] = eval_dataset
        return cls(datasets=datasets, dataloader_creators=dataloader_creators)


def create_train_eval_datasets_v1(
    train_path: Path | str | None,
    eval_path: Path | str | None,
    input_size: int = 224,
) -> tuple[Dataset | None, Dataset | None]:
    r"""Creates the training and evaluation datasets with the original
    data augmentations (V1).

    This function initializes the original data augmentations,
    sometimes called V1
    (https://pytorch.org/blog/introducing-torchvision-new-multi-weight-support-api/).
    Note that more aggressive data augmentation can be used to
    improve the performances. See the following article for more
    information:

        Wightman R., Touvron H., Jégou H.
        ResNet strikes back: An improved training procedure in timm.
        http://arxiv.org/pdf/2110.00476

    Args:
    ----
        train_path (``pathlib.Path`` or str or ``None``): Specifies
            the path to the training dataset. If ``None``, the
            training dataset is set to ``None``.
        eval_path (``pathlib.Path`` or str or ``None``): Specifies
            the path to the evaluation dataset. If ``None``, the
            evaluation dataset is set to ``None``.
        input_size (int, optional): Specifies the input size for the
            image. Default: ``224``

    Returns:
    -------
        (``Dataset`` or ``None``, ``Dataset`` or ``None``):
            The training and evaluation datasets. The datasets use the
            original image data augmentation (V1).
    """
    check_torchvision()
    normalize = transforms.Normalize(mean=IMAGENET_DEFAULT_MEAN, std=IMAGENET_DEFAULT_STD)
    train_dataset, eval_dataset = None, None
    if train_path:
        logger.info(f"Initialize the training dataset from {train_path}")
        train_dataset = ImageFolderDataset(
            train_path,
            transforms.Compose(
                [
                    transforms.RandomResizedCrop(input_size),
                    transforms.RandomHorizontalFlip(),
                    transforms.ToTensor(),
                    normalize,
                ]
            ),
        )
        logger.info(f"number of training examples: {len(train_dataset):,}")

    if eval_path:
        logger.info(f"Initialize the evaluation dataset from {eval_path}")
        eval_dataset = ImageFolderDataset(
            eval_path,
            transforms.Compose(
                [
                    # Maintain same ratio w.r.t. 256 -> 224 images
                    transforms.Resize(int((256 / 224) * input_size)),
                    transforms.CenterCrop(input_size),
                    transforms.ToTensor(),
                    normalize,
                ]
            ),
        )
        logger.info(f"number of evaluation examples: {len(eval_dataset):,}")
    return train_dataset, eval_dataset



from __future__ import annotations

__all__ = ["EmptyMeterError"]


class EmptyMeterError(Exception):
    r"""Generates an error if the metric is empty because it is not
    possible to evaluate an empty metric."""



r"""The code is in this module is designed to be used for testing
purpose only."""

from __future__ import annotations

__all__ = [
    "DummyDataset",
    "DummyIterableDataset",
    "DummyClassificationModel",
    "DummyDataSource",
    "create_dummy_engine",
]

from collections.abc import Iterator

import torch
from objectory import OBJECT_TARGET
from torch.nn import CrossEntropyLoss, Linear, Module
from torch.optim import Optimizer
from torch.utils.data import Dataset, IterableDataset

from gravitorch import constants as ct
from gravitorch.datasources import BaseDataSource, DatasetDataSource
from gravitorch.engines import BaseEngine
from gravitorch.models import BaseModel


class DummyDataset(Dataset):
    r"""Implements a dummy map-style dataset for testing purpose.

    Args:
    ----
        feature_size (dim, optional): Specifies the feature size.
            Default: ``4``
        num_examples (dim, optional): Specifies the number of
            examples. Default: ``8``
    """

    def __init__(self, feature_size: int = 4, num_examples: int = 8) -> None:
        self._feature_size = int(feature_size)
        self._num_examples = int(num_examples)

    def __getitem__(self, item: int) -> dict:
        return {ct.INPUT: torch.ones(self._feature_size) + item, ct.TARGET: 1}

    def __len__(self) -> int:
        return self._num_examples

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(num_examples={self._feature_size:,}, "
            f"feature_size={self._feature_size:,})"
        )


class DummyIterableDataset(IterableDataset):
    r"""Implements a dummy iterable-style dataset for testing purpose.

    Args:
    ----
        feature_size (dim, optional): Specifies the feature size.
            Default: ``4``
        num_examples (dim, optional): Specifies the number of
            examples. Default: ``8``
        has_length (bool, optional): If ``True``, the length of the
            dataset is defined, otherwise it returns ``TypeError``.
            Default: ``False``
    """

    def __init__(
        self, feature_size: int = 4, num_examples: int = 8, has_length: bool = False
    ) -> None:
        self._feature_size = int(feature_size)
        self._num_examples = int(num_examples)
        self._has_length = bool(has_length)
        self._iteration = 0

    def __iter__(self) -> Iterator:
        self._iteration = 0
        return self

    def __next__(self) -> dict:
        self._iteration += 1
        if self._iteration > self._num_examples:
            raise StopIteration

        return {ct.INPUT: torch.ones(self._feature_size) + self._iteration, ct.TARGET: 1}

    def __len__(self) -> int:
        if self._has_length:
            return self._num_examples
        raise TypeError(f"{type(self).__qualname__} instance doesn't have valid length")

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(num_examples={self._feature_size:,}, "
            f"feature_size={self._feature_size:,})"
        )


class DummyDataSource(DatasetDataSource):
    r"""Implements a dummy datasource for testing purpose.

    Args:
    ----
        train_dataset (``Dataset`` or ``None``, optional): Specifies
            the training dataset. If ``None``, a dummy map-style
            dataset is automatically created. Default: ``None``
        eval_dataset (``Dataset`` or ``None``, optional): Specifies
            the evaluation dataset. If ``None``, a dummy map-style
            dataset is automatically created. Default: ``None``
        batch_size (int, optional): Specifies the batch size.
            Default: ``1``
    """

    def __init__(
        self,
        train_dataset: Dataset | None = None,
        eval_dataset: Dataset | None = None,
        batch_size: int | None = 1,
    ) -> None:
        if train_dataset is None:
            train_dataset = DummyDataset()
        if eval_dataset is None:
            eval_dataset = DummyDataset()

        from gravitorch.creators.dataloader import DataLoaderCreator

        super().__init__(
            datasets={ct.TRAIN: train_dataset, ct.EVAL: eval_dataset},
            dataloader_creators={
                ct.TRAIN: DataLoaderCreator(batch_size=batch_size, shuffle=False),
                ct.EVAL: DataLoaderCreator(batch_size=batch_size, shuffle=False),
            },
        )


class DummyClassificationModel(BaseModel):
    r"""Implements a dummy classification model for testing purpose.

    Args:
    ----
        feature_size (dim, optional): Specifies the feature size.
            Default: ``4``
        num_classes (dim, optional): Specifies the number of classes.
            Default: ``3``
        loss_nan (bool, optional): If ``True``, the forward function
            returns a loss filled with a NaN value.
    """

    def __init__(self, feature_size: int = 4, num_classes: int = 3, loss_nan: bool = False) -> None:
        super().__init__()
        self.linear = Linear(feature_size, num_classes)
        self.criterion = CrossEntropyLoss()
        self._return_loss_nan = bool(loss_nan)

    def forward(self, batch: dict) -> dict:
        if self._return_loss_nan:
            return {ct.LOSS: torch.tensor(float("nan"))}
        return {ct.LOSS: self.criterion(self.linear(batch[ct.INPUT]), batch[ct.TARGET])}


def create_dummy_engine(
    datasource: BaseDataSource | dict | None = None,
    model: Module | dict | None = None,
    optimizer: Optimizer | dict | None = None,
    device: torch.device | None = None,
    **kwargs,
) -> BaseEngine:
    r"""Creates an engine with dummy components for testing purpose.

    Args:
    ----
        datasource (``BaseDataSource`` or dict or ``None``): Specifies
            the datasource or its configuration. If ``None``, a dummy
            datasource is automatically created. Default: ``None``
        model (``Module`` or dict or ``None``): Specifies the model or
            its configuration. If ``None``, a dummy classification model
            is automatically created. Default: ``None``
        model (``Optimizer`` or dict or ``None``): Specifies the
            optimizer or its configuration. If ``None``, a SGD
            optimizer is automatically created. Default: ``None``
        device (``torch.device`` or ``None``): Specifies the target
            device. Default: ``None``
        **kwargs: Arbitrary keyword arguments.
    """
    datasource = datasource or DummyDataSource(batch_size=2)
    model = model or DummyClassificationModel()
    optimizer = optimizer or {OBJECT_TARGET: "torch.optim.SGD", "lr": 0.01}

    from gravitorch.creators.core import VanillaCoreCreator
    from gravitorch.engines import AlphaEngine

    return AlphaEngine(
        core_creator=VanillaCoreCreator(
            datasource=datasource,
            model=model.to(device=device),
            optimizer=optimizer,
        ),
        **kwargs,
    )



from __future__ import annotations

__all__ = ["IterableDataFlow"]

import copy
import logging
from collections.abc import Iterable, Iterator
from typing import TypeVar

from gravitorch.experimental.dataflow.base import BaseDataFlow

logger = logging.getLogger(__name__)

T = TypeVar("T")


class IterableDataFlow(BaseDataFlow[T]):
    r"""Implements a simple dataflow for iterables.

    Args:
    ----
        iterable (``Iterable``): Specifies the iterable.
        deepcopy (bool, optional): If ``True``, the input iterable
            object is deep-copied before to iterate over the data.
            It allows a deterministic behavior when in-place
            operations are performed on the data. Default: ``False``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.experimental.dataflow import IterableDataFlow
        >>> with IterableDataFlow([1, 2, 3, 4, 5]) as dataflow:
        ...     for batch in dataflow:
        ...         print(batch)  # do something
        ...
    """

    def __init__(self, iterable: Iterable[T], deepcopy: bool = False) -> None:
        if not isinstance(iterable, Iterable):
            raise TypeError(f"Incorrect type. Expecting iterable but received {type(iterable)}")
        self.iterable = iterable
        self._deepcopy = bool(deepcopy)

    def __iter__(self) -> Iterator[T]:
        iterable = self.iterable
        if self._deepcopy:  # TODO: move to launch
            logger.info("Copying the input iterable...")
            try:
                iterable = copy.deepcopy(iterable)
            except TypeError:
                logger.warning(
                    "The input iterable can not be deepcopied, please be aware of in-place "
                    "modification would affect source data."
                )
        yield from iterable

    def __repr__(self) -> str:
        try:
            extra = f"length={len(self.iterable):,}"
        except TypeError:
            extra = ""
        return f"{self.__class__.__qualname__}({extra})"

    def launch(self) -> None:
        r"""Nothing to do for this dataflow."""

    def shutdown(self) -> None:
        r"""Nothing to do for this dataflow."""



__all__ = [
    "Constant",
    "ConstantBias",
    "constant",
    "constant_bias",
]

import logging
from typing import Union

from torch import nn
from torch.nn import Module

from gravitorch.nn.init.base import BaseInitializer

logger = logging.getLogger(__name__)


class Constant(BaseInitializer):
    r"""Implements a module parameter initializer where the weights are
    initialized with constant values.

    Args:
    ----
        value (float): Specifies the value to initialize the
            parameters with.
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``
        log_info (bool, optional): If ``True``, log some information
            about the weights that are initialized. Default: ``False``
    """

    def __init__(
        self,
        value: Union[int, float] = 0.0,
        learnable_only: bool = True,
        log_info: bool = False,
    ) -> None:
        super().__init__()
        self._value = float(value)
        self._learnable_only = bool(learnable_only)
        self._log_info = bool(log_info)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(value={self._value}, "
            f"learnable_only={self._learnable_only}, log_info={self._log_info})"
        )

    def initialize(self, module: Module) -> None:
        logger.info(
            f"Initializing weights with {self._value} (learnable_only: {self._learnable_only})..."
        )
        constant(
            module=module,
            value=self._value,
            learnable_only=self._learnable_only,
            log_info=self._log_info,
        )


class ConstantBias(BaseInitializer):
    r"""Implements a module parameter initializer where the biases are
    initialized with constant values.

    Args:
    ----
        value (float): Specifies the value to initialize the
            parameters with.
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``
        log_info (bool, optional): If ``True``, log some information
            about the biases that are initialized. Default: ``False``
    """

    def __init__(
        self,
        value: Union[int, float] = 0.0,
        learnable_only: bool = True,
        log_info: bool = False,
    ) -> None:
        super().__init__()
        self._value = float(value)
        self._learnable_only = bool(learnable_only)
        self._log_info = bool(log_info)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(value={self._value}, "
            f"learnable_only={self._learnable_only}, log_info={self._log_info})"
        )

    def initialize(self, module: Module) -> None:
        logger.info(
            f"Initializing biases with {self._value} (learnable_only: {self._learnable_only})..."
        )
        constant_bias(
            module=module,
            value=self._value,
            learnable_only=self._learnable_only,
            log_info=self._log_info,
        )


def constant_bias(
    module: Module,
    value: Union[int, float],
    learnable_only: bool = True,
    log_info: bool = False,
) -> None:
    r"""Recursively initialize the biases with ``value``.

    To identify the biases, this function looks at if ``'bias'`` is
    in the parameter name.

    Args:
    ----
        module (``torch.nn.Module``): Specifies the module to
            initialize.
        value (float): Specifies the value to initialize the
            parameters with.
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``
        log_info (bool, optional): If ``True``, log some information
            about the weights that are initialized. Default: ``False``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.nn.init import constant_bias
        >>> from torch import nn
        >>> net = nn.Sequential(nn.Linear(4, 6), nn.ReLU(), nn.BatchNorm1d(6), nn.Linear(6, 1))
        >>> constant_bias(net, 2)
        >>> for key, param in net.named_parameters():
        ...     print(key, param)
        ...
        0.weight Parameter containing:
        tensor([[ 0.1911,  0.2128, -0.3738,  0.3777],
                [ 0.3394,  0.1795, -0.1418,  0.1076],
                [-0.2050, -0.0029, -0.2555,  0.2612],
                [ 0.1260, -0.3503,  0.1254, -0.3669],
                [ 0.1261, -0.1652,  0.1695,  0.3879],
                [-0.0405,  0.0372,  0.0515, -0.4335]], requires_grad=True)
        0.bias Parameter containing:
        tensor([2., 2., 2., 2., 2., 2.], requires_grad=True)
        2.weight Parameter containing:
        tensor([1., 1., 1., 1., 1., 1.], requires_grad=True)
        2.bias Parameter containing:
        tensor([2., 2., 2., 2., 2., 2.], requires_grad=True)
        3.weight Parameter containing:
        tensor([[-0.2702,  0.2759, -0.2408,  0.1202, -0.1770,  0.0223]],
               requires_grad=True)
        3.bias Parameter containing:
        tensor([2.], requires_grad=True)
    """
    for name, params in module.named_parameters():
        if "bias" in name and (not learnable_only or learnable_only and params.requires_grad):
            if log_info:
                logger.info(f"Initializing bias '{name}' with {value} | shape={params.shape}")
            nn.init.constant_(params.data, value)


def constant(
    module: Module,
    value: Union[int, float],
    learnable_only: bool = True,
    log_info: bool = False,
) -> None:
    r"""Initializes the parameters of the module with a constant
    ``value``.

    Args:
    ----
        module (``torch.nn.Module``): Specifies the module with the
            parameters to initialize.
        value (int, float): Specifies the value to initialize the
            parameters with.
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``
        log_info (bool, optional): If ``True``, log some information
            about the weights that are initialized. Default: ``False``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.nn.init import constant
        >>> from torch import nn
        >>> net = nn.Sequential(nn.Linear(4, 6), nn.ReLU(), nn.BatchNorm1d(6), nn.Linear(6, 1))
        >>> constant(net, 2.0)
        >>> for key, param in net.named_parameters():
        ...     print(key, param)
        ...
        0.weight Parameter containing:
        tensor([[2., 2., 2., 2.],
                [2., 2., 2., 2.],
                [2., 2., 2., 2.],
                [2., 2., 2., 2.],
                [2., 2., 2., 2.],
                [2., 2., 2., 2.]], requires_grad=True)
        0.bias Parameter containing:
        tensor([2., 2., 2., 2., 2., 2.], requires_grad=True)
        2.weight Parameter containing:
        tensor([2., 2., 2., 2., 2., 2.], requires_grad=True)
        2.bias Parameter containing:
        tensor([2., 2., 2., 2., 2., 2.], requires_grad=True)
        3.weight Parameter containing:
        tensor([[2., 2., 2., 2., 2., 2.]], requires_grad=True)
        3.bias Parameter containing:
        tensor([2.], requires_grad=True)
    """
    for name, params in module.named_parameters():
        if not learnable_only or learnable_only and params.requires_grad:
            if log_info:
                logger.info(f"Initializing '{name}' with {value} | shape={params.shape}")
            nn.init.constant_(params.data, value)



r"""This package contains the implementation of some data loader
creators."""

__all__ = [
    "AutoDataLoaderCreator",
    "BaseDataLoaderCreator",
    "DistributedDataLoaderCreator",
    "DataLoaderCreator",
    "setup_dataloader_creator",
]

from gravitorch.creators.dataloader.base import BaseDataLoaderCreator
from gravitorch.creators.dataloader.factory import setup_dataloader_creator
from gravitorch.creators.dataloader.pytorch import (
    AutoDataLoaderCreator,
    DataLoaderCreator,
    DistributedDataLoaderCreator,
)



from __future__ import annotations

__all__ = ["AdvancedCoreCreator"]


from torch.nn import Module
from torch.optim import Optimizer

from gravitorch.creators.core.base import BaseCoreCreator
from gravitorch.creators.datasource.base import (
    BaseDataSourceCreator,
    setup_datasource_creator,
)
from gravitorch.creators.lr_scheduler.base import BaseLRSchedulerCreator
from gravitorch.creators.lr_scheduler.utils import setup_lr_scheduler_creator
from gravitorch.creators.model.base import BaseModelCreator, setup_model_creator
from gravitorch.creators.optimizer.base import BaseOptimizerCreator
from gravitorch.creators.optimizer.utils import setup_optimizer_creator
from gravitorch.datasources.base import BaseDataSource
from gravitorch.engines.base import BaseEngine
from gravitorch.lr_schedulers.base import LRSchedulerType
from gravitorch.utils.format import str_indent


class AdvancedCoreCreator(BaseCoreCreator):
    r"""Implements an advanced core engine moules creator.

    Args:
    ----
        datasource_creator (``BaseDataSourceCreator`` or dict):
            Specifies the datasource creator or its configuration.
        model_creator (``BaseModelCreator`` or dict): Specifies the
            model creator or its configuration.
        optimizer_creator (``BaseOptimizerCreator`` or dict or
            ``None`): Specifies the optimizer creator or its
            configuration. Default: ``None``
        lr_scheduler_creator (``BaseLRSchedulerCreator`` or dict or
            ``None`): Specifies the LR scheduler creator or its
            configuration. Default: ``None``
    """

    def __init__(
        self,
        datasource_creator: BaseDataSourceCreator | dict,
        model_creator: BaseModelCreator | dict,
        optimizer_creator: BaseOptimizerCreator | dict | None = None,
        lr_scheduler_creator: BaseLRSchedulerCreator | dict | None = None,
    ) -> None:
        self._datasource_creator = setup_datasource_creator(datasource_creator)
        self._model_creator = setup_model_creator(model_creator)
        self._optimizer_creator = setup_optimizer_creator(optimizer_creator)
        self._lr_scheduler_creator = setup_lr_scheduler_creator(lr_scheduler_creator)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  datasource_creator={str_indent(self._datasource_creator)},\n"
            f"  model_creator={str_indent(self._model_creator)},\n"
            f"  optimizer_creator={str_indent(self._optimizer_creator)},\n"
            f"  lr_scheduler_creator={str_indent(self._lr_scheduler_creator)},\n"
            ")"
        )

    def create(
        self, engine: BaseEngine
    ) -> tuple[BaseDataSource, Module, Optimizer | None, LRSchedulerType | None]:
        datasource = self._datasource_creator.create(engine=engine)
        model = self._model_creator.create(engine=engine)
        optimizer = self._optimizer_creator.create(engine=engine, model=model)
        lr_scheduler = self._lr_scheduler_creator.create(engine=engine, optimizer=optimizer)
        return datasource, model, optimizer, lr_scheduler



r"""This module defines a criterion wrapper to make PyTorch criteria
compatible with ``gravitorch.models.VanillaModel`` and packed
sequences."""

from __future__ import annotations

__all__ = ["PackedSequenceLoss"]


from torch import Tensor
from torch.nn import Module

from gravitorch import constants as ct
from gravitorch.nn import setup_module


class PackedSequenceLoss(Module):
    r"""Implements a wrapper to adapt PyTorch criterion to deal with
    packed sequence inputs.

    This class works for every PyTorch criterion that has two inputs:
    prediction and target.

    Args:
    ----
        criterion (``torch.nn.Module`` or dict): Specifies the loss
            function or its configuration.
        prediction_key (str): Specifies the prediction key.
            Default: ``"prediction"``.
        target_key (str): Specifies the target key.
            Default: ``"target"``.
        mask_key (str): Specifies the mask key.
            Default: ``"mask"``.

    Example usage:

    .. code-block:: pycon

        >>> from torch import nn
        >>> from gravitorch.models.criteria import PackedSequenceLoss
        # Init with a nn.Module
        >>> criterion = PackedSequenceLoss(criterion=nn.MSELoss())
        # Init with a config
        >>> criterion = PackedSequenceLoss(criterion={"_target_": "torch.nn.MSELoss"})
        # Customize keys.
        >>> criterion = PackedSequenceLoss(
        ...     criterion=nn.MSELoss(),
        ...     prediction_key="my_prediction",
        ...     target_key="my_target",
        ...     mask_key="my_mask",
        ... )
        >>> net_out = {"my_prediction": ...}
        >>> batch = {"my_target": ..., "my_mask": ...}
        >>> loss = criterion(net_out, batch)
        {'loss': torch.tensor(...)}
    """

    def __init__(
        self,
        criterion: Module | dict,
        prediction_key: str = ct.PREDICTION,
        target_key: str = ct.TARGET,
        mask_key: str = ct.MASK,
    ) -> None:
        super().__init__()
        self.criterion = setup_module(criterion)
        self._prediction_key = prediction_key
        self._target_key = target_key
        self._mask_key = mask_key

    def forward(self, net_out: dict, batch: dict) -> dict[str, Tensor]:
        r"""Computes the loss value given the network output and the
        batch.

        Args:
        ----
            net_out (dict): Specifies the network output which
                contains the prediction.
            batch (dict): Specifies the batch which contains the
                target and the mask.

        Returns:
        -------
            dict: a dict with the loss value.
        """
        prediction = self._get_prediction_from_net_out(net_out)
        target = self._get_target_from_batch(batch)

        # Get the mask and remove the examples that are 0-masked.
        mask = batch.get(self._mask_key, None)
        if mask is not None:
            prediction = prediction[mask.data > 0]
            target = target[mask.data > 0]

        return {ct.LOSS: self.criterion(prediction, target)}

    def _get_prediction_from_net_out(self, net_out: dict) -> Tensor:
        r"""Gets the prediction from the network output.

        Args:
        ----
            net_out (dict): Specifies the network output
                which contains the prediction.

        Returns:
        -------
            ``torch.Tensor``: the prediction
        """
        return net_out[self._prediction_key].data

    def _get_target_from_batch(self, batch: dict) -> Tensor:
        r"""Gets the target from the batch. The target is the tensor with
        the key 'target'.

        Args:
        ----
            batch (dict): Specifies the batch which contains the target.

        Returns:
        -------
            ``torch.Tensor``: the target
        """
        return batch[self._target_key].data



__all__ = ["BaseInitializer"]

from abc import ABC, abstractmethod

from objectory import AbstractFactory
from torch.nn import Module


class BaseInitializer(ABC, metaclass=AbstractFactory):
    r"""Defines the base parameter initializer.

    Note that there are other ways to initialize the model parameters.
    For example, you can initialize the weights of your model directly
    in the model.
    """

    @abstractmethod
    def initialize(self, module: Module) -> None:
        r"""Initializes the parameters of the model.

        The parameters of the model should be updated in-place.

        Args:
        ----
            module (``torch.nn.Module``): Specifies the module to
                initialize.
        """



from __future__ import annotations

__all__ = ["CacheDataCreator"]

import copy
import logging
from typing import Any, TypeVar

from gravitorch.data.datacreators.base import BaseDataCreator, setup_data_creator
from gravitorch.engines.base import BaseEngine
from gravitorch.utils.format import str_indent

logger = logging.getLogger(__name__)

T = TypeVar("T")


class CacheDataCreator(BaseDataCreator[T]):
    r"""Implements a data creator that creates the data only once, then
    cache them and return them.

    Args:
    ----
        creator (``BaseDataCreator`` or dict): Specifies a data
            creator or its configuration.
        deepcopy (bool, optional): If ``True``, a deepcopy of the data
            is done before to return the data. If ``deepcopy`` is
            explicitly set to ``False``, users should ensure that
            the data pipeline does not contain any in-place operations
            over the data to prevent data inconsistency if the
            ``create`` method is called multiple times.
    """

    def __init__(self, creator: BaseDataCreator[T] | dict, deepcopy: bool = False) -> None:
        self._creator = setup_data_creator(creator)
        self._deepcopy = bool(deepcopy)
        self._is_cache_created = False
        # This variable is used to cache the data. The type depends on the value returned by
        # the function ``create`` of the data creator object.
        self._cached_data: Any = None

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(\n"
            f"  creator={str_indent(self._creator)},\n"
            f"  is_cache_created={self._is_cache_created},\n"
            f"  deepcopy={self._deepcopy},\n"
            ")"
        )

    @property
    def creator(self) -> BaseDataCreator[T]:
        r"""``BaseDataCreator``: The data creator."""
        return self._creator

    @property
    def deepcopy(self) -> bool:
        r"""``bool``: Indicates if a deepcopy of the data is done before
        to return the data."""
        return self._deepcopy

    def create(self, engine: BaseEngine | None = None) -> T:
        r"""Creates data.

        Args:
        ----
            engine (``BaseEngine`` or ``None``): Specifies an engine.
                Default: ``None``

        Return:
        ------
            The created data. The returned data depends on the data
                creator.
        """
        if not self._is_cache_created:
            logger.info("Creating data and caching them...")
            self._cached_data = self._creator.create(engine)
            self._is_cache_created = True
        data = self._cached_data
        if self._deepcopy:
            try:
                data = copy.deepcopy(data)
            except TypeError:
                logger.warning(
                    "The data can not be deepcopied. "
                    "Please be aware of in-place modification would affect source data"
                )
        return data



r"""This package contains the implementation of some data loader
collators."""

from __future__ import annotations

__all__ = [
    "BaseCollator",
    "DictPackedSequenceCollator",
    "DictPaddedSequenceCollator",
    "PackedSequenceCollator",
    "PaddedSequenceCollator",
    "setup_collator",
]

from gravitorch.data.dataloaders.collators.base import BaseCollator, setup_collator
from gravitorch.data.dataloaders.collators.pack import (
    DictPackedSequenceCollator,
    PackedSequenceCollator,
)
from gravitorch.data.dataloaders.collators.pad import (
    DictPaddedSequenceCollator,
    PaddedSequenceCollator,
)



from __future__ import annotations

__all__ = [
    "create_dataloader",
    "create_dataloader2",
    "is_dataloader_config",
    "is_dataloader2_config",
    "setup_dataloader",
    "setup_dataloader2",
]

import logging
from collections.abc import Iterable

from objectory import factory
from objectory.utils import is_object_config
from torch.utils.data import DataLoader, Dataset
from torch.utils.data.graph import DataPipe

from gravitorch.data.datasets.factory import setup_dataset
from gravitorch.utils.factory import setup_object
from gravitorch.utils.format import str_target_object
from gravitorch.utils.imports import check_torchdata, is_torchdata_available

if is_torchdata_available():
    from torchdata.dataloader2 import DataLoader2, ReadingServiceInterface
    from torchdata.dataloader2.adapter import Adapter
else:  # pragma: no cover
    Adapter = "Adapter"
    DataLoader2 = "DataLoader2"
    ReadingServiceInterface = "ReadingServiceInterface"

logger = logging.getLogger(__name__)


def create_dataloader(dataset: Dataset | dict, **kwargs) -> DataLoader:
    r"""Instantiates a ``torch.utils.data.DataLoader`` from a
    ``torch.utils.data.Dataset`` or its configuration.

    Args:
        dataset (``torch.utils.data.Dataset`` or ``dict``): Specifies
            a dataset or its configuration.
        **kwargs: See ``torch.utils.data.DataLoader`` documentation.

    Returns:
        ``torch.utils.data.DataLoader``: The instantiated dataloader.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.dataloaders import create_dataloader
        >>> create_dataloader(
        ...     {
        ...         "_target_": "gravitorch.data.datasets.DummyMultiClassDataset",
        ...         "num_examples": 10,
        ...         "num_classes": 2,
        ...         "feature_size": 4,
        ...     }
        ... )
        <torch.utils.data.dataloader.DataLoader at 0x0123456789>
    """
    return DataLoader(
        setup_dataset(dataset), **{key: setup_object(value) for key, value in kwargs.items()}
    )


def create_dataloader2(
    datapipe: DataPipe | dict,
    datapipe_adapter_fn: Iterable[Adapter | dict] | Adapter | dict | None = None,
    reading_service: ReadingServiceInterface | dict | None = None,
) -> DataLoader2:
    r"""Instantiates a ``torchdata.dataloader2.DataLoader2`` object from
    a ``DataPipe`` or its configuration.

    Args:
        datapipe: Specifies the DataPipe or its configuration.
        datapipe_adapter_fn: Specifies the ``Adapter`` function(s) that
            will be applied to the DataPipe. Default: ``None``
        reading_service: Defines how ``DataLoader2`` should execute
            operations over the ``DataPipe``, e.g.
            multiprocessing/distributed. Default: ``None``

    Returns:
        ``torchdata.dataloader2.DataLoader2``: The instantiated
            ``DataLoader2``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.dataloaders import create_dataloader2
        >>> create_dataloader2(
        ...     {
        ...         "_target_": "torch.utils.data.datapipes.iter.IterableWrapper",
        ...         "iterable": range(10),
        ...     }
        ... )
        <torchdata.dataloader2.dataloader2.DataLoader2 at 0x0123456789>
    """
    check_torchdata()
    return DataLoader2(
        datapipe=setup_object(datapipe),
        datapipe_adapter_fn=setup_object(datapipe_adapter_fn),
        reading_service=setup_object(reading_service),
    )


def is_dataloader_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``torch.utils.data.DataLoader``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``torch.utils.data.DataLoader`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.dataloaders import is_dataloader_config
        >>> is_dataloader_config({"_target_": "torch.utils.data.DataLoader"})
        True
    """
    return is_object_config(config, DataLoader)


def is_dataloader2_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``torchdata.dataloader2.DataLoader2``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``torchdata.dataloader2.DataLoader2`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.dataloaders import is_dataloader2_config
        >>> is_dataloader2_config({"_target_": "torchdata.dataloader2.DataLoader2"})
        True
    """
    check_torchdata()
    return is_object_config(config, DataLoader2)


def setup_dataloader(dataloader: DataLoader | dict) -> DataLoader:
    r"""Sets up a ``torch.utils.data.DataLoader`` object.

    Args:
    ----
        dataloader (``torch.utils.data.DataLoader`` or dict):
            Specifies the dataloader or its configuration (dictionary).

    Returns:
    -------
        ``torch.utils.data.DataLoader``: The instantiated dataloader.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.dataloaders import setup_dataloader
        >>> from gravitorch.data.datasets import ExampleDataset
        >>> dataloader = setup_dataloader(
        ...     {"_target_": "torch.utils.data.DataLoader", "dataset": ExampleDataset((1, 2, 3, 4))}
        ... )
        >>> dataloader
        <torch.utils.data.dataloader.DataLoader at 0x119bd42e0>
        >>> setup_dataloader(
        ...     dataloader
        ... )  # Do nothing because the dataloader is already instantiated
        <torch.utils.data.dataloader.DataLoader at 0x119bd42e0>
    """
    if isinstance(dataloader, dict):
        logger.info(
            "Initializing a `torch.utils.data.DataLoader` from its configuration... "
            f"{str_target_object(dataloader)}"
        )
        dataloader = factory(**dataloader)
    return dataloader


def setup_dataloader2(dataloader: DataLoader2 | dict) -> DataLoader2:
    r"""Sets up a ``torchdata.dataloader2.DataLoader2`` object.

    Args:
    ----
        dataloader (``torchdata.dataloader2.DataLoader2`` or dict):
            Specifies the dataloader or its configuration (dictionary).

    Returns:
    -------
        ``torchdata.dataloader2.DataLoader2``: The instantiated dataloader.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.dataloaders import setup_dataloader2
        >>> from torch.utils.data.datapipes.iter import IterableWrapper
        >>> dataloader = setup_dataloader2(
        ...     {
        ...         "_target_": "torchdata.dataloader2.DataLoader2",
        ...         "datapipe": IterableWrapper((1, 2, 3, 4)),
        ...     }
        ... )
        >>> dataloader
        <torch.utils.data.dataloader.DataLoader at 0x119bd42e0>
        >>> setup_dataloader(
        ...     dataloader
        ... )  # Do nothing because the dataloader is already instantiated
        <torch.utils.data.dataloader.DataLoader at 0x119bd42e0>
    """
    check_torchdata()
    if isinstance(dataloader, dict):
        logger.info(
            "Initializing a `torchdata.dataloader2.DataLoader2` from its configuration... "
            f"{str_target_object(dataloader)}"
        )
        dataloader = factory(**dataloader)
    return dataloader



from __future__ import annotations

__all__ = ["is_map_datapipe_config", "setup_map_datapipe"]

import logging

from objectory import factory
from objectory.utils import is_object_config
from torch.utils.data import MapDataPipe

from gravitorch.utils.format import str_target_object

logger = logging.getLogger(__name__)


def is_map_datapipe_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``MapDataPipe``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
    ----
        config (dict): Specifies the configuration to check.

    Returns:
    -------
        bool: ``True`` if the input configuration is a configuration
            for a ``MapDataPipe`` object.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.datapipes import is_map_datapipe_config
        >>> is_map_datapipe_config(
        ...     {
        ...         "_target_": "torch.utils.data.datapipes.map.SequenceWrapper",
        ...         "sequence": [1, 2, 3, 4],
        ...     }
        ... )
        True
    """
    return is_object_config(config, MapDataPipe)


def setup_map_datapipe(datapipe: MapDataPipe | dict) -> MapDataPipe:
    r"""Set up an ``MapDataPipe``.

    Args:
    ----
        datapipe (``MapDataPipe`` or dict): Specifies a ``MapDataPipe``
            or its configuration.

    Returns:
    -------
        ``MapDataPipe``: The instantiated ``MapDataPipe``.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.datapipes import setup_map_datapipe
        >>> datapipe = setup_map_datapipe(
        ...     {
        ...         "_target_": "torch.utils.data.datapipes.map.SequenceWrapper",
        ...         "sequence": [1, 2, 3, 4],
        ...     }
        ... )
        >>> tuple(datapipe)
        (1, 2, 3, 4)
    """
    if isinstance(datapipe, dict):
        logger.info(
            f"Initializing a `MapDataPipe` from its configuration... {str_target_object(datapipe)}"
        )
        datapipe = factory(**datapipe)
    return datapipe



__all__ = [
    "BaseKaiming",
    "KaimingNormal",
    "KaimingUniform",
    "kaiming_normal",
    "kaiming_uniform",
]

import logging

from torch import nn
from torch.nn import Module

from gravitorch.nn.init.base import BaseInitializer

logger = logging.getLogger(__name__)


class BaseKaiming(BaseInitializer):
    r"""Implements a module parameter initializer with the Kaiming Normal
    or uniform strategy.

    Args:
    ----
        neg_slope (float, optional): Specifies the negative slope of
            the rectifier used after this layer (only used with
            ``'leaky_relu'``). Default: ``0.0``
        mode (str, optional): either ``'fan_in'`` or ``'fan_out'``.
            Choosing ``'fan_in'`` preserves the magnitude of the
            variance of the weights in the forward pass. Choosing
            ``'fan_out'`` preserves the magnitudes in the backwards
            pass. Default: ``'fan_in'``
        nonlinearity (str, optional): the non-linear function
            (`nn.functional` name), recommended to use only with
            ``'relu'`` or ``'leaky_relu'``. Default: ``'leaky_relu'``
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``
        log_info (bool, optional): If ``True``, log some information
            about the weights that are initialized. Default: ``False``
    """

    def __init__(
        self,
        neg_slope: float = 0.0,
        mode: str = "fan_in",
        nonlinearity: str = "leaky_relu",
        learnable_only: bool = True,
        log_info: bool = False,
    ) -> None:
        super().__init__()
        self._neg_slope = float(neg_slope)
        self._mode = str(mode)
        self._nonlinearity = str(nonlinearity)
        self._learnable_only = bool(learnable_only)
        self._log_info = bool(log_info)

    def __str__(self) -> str:
        return (
            f"{self.__class__.__qualname__}(mode={self._mode}, nonlinearity={self._nonlinearity}, "
            f"neg_slope={self._neg_slope}, learnable_only={self._learnable_only}, "
            f"log_info={self._log_info})"
        )


class KaimingNormal(BaseKaiming):
    r"""Implements a module parameter initializer with the Kaiming
    Normal strategy."""

    def initialize(self, module: Module) -> None:
        logger.info(
            f"Initializing module parameters with the Kaiming Normal strategy "
            f"(mode={self._mode}, nonlinearity={self._nonlinearity}, neg_slope={self._neg_slope}, "
            f"learnable_only={self._learnable_only})..."
        )
        kaiming_normal(
            module=module,
            neg_slope=self._neg_slope,
            mode=self._mode,
            nonlinearity=self._nonlinearity,
            learnable_only=self._learnable_only,
            log_info=self._log_info,
        )


class KaimingUniform(BaseKaiming):
    r"""Implements a module parameter initializer with the Kaiming
    uniform strategy."""

    def initialize(self, module: Module) -> None:
        logger.info(
            f"Initializing module parameters with the Kaiming uniform strategy "
            f"(mode={self._mode}, nonlinearity={self._nonlinearity}, neg_slope={self._neg_slope}, "
            f"learnable_only={self._learnable_only})..."
        )
        kaiming_uniform(
            module=module,
            neg_slope=self._neg_slope,
            mode=self._mode,
            nonlinearity=self._nonlinearity,
            learnable_only=self._learnable_only,
            log_info=self._log_info,
        )


def kaiming_normal(
    module: nn.Module,
    neg_slope: float = 0.0,
    mode: str = "fan_in",
    nonlinearity: str = "leaky_relu",
    learnable_only: bool = True,
    log_info: bool = False,
) -> None:
    r"""Initializes the module parameters with the Kaiming Normal
    strategy.

    Args:
    ----
        module (``torch.nn.Module``): Specifies the module to
            initialize.
        neg_slope (float, optional): Specifies the negative slope of
            the rectifier used after this layer (only used with
            ``'leaky_relu'``). Default: ``0.0``
        mode (str, optional): either ``'fan_in'`` or ``'fan_out'``.
            Choosing ``'fan_in'`` preserves the magnitude of the
            variance of the weights in the forward pass. Choosing
            ``'fan_out'`` preserves the magnitudes in the backwards
            pass. Default: ``'fan_in'``
        nonlinearity (str, optional): the non-linear function
            (`nn.functional` name), recommended to use only with
            ``'relu'`` or ``'leaky_relu'``. Default: ``'leaky_relu'``
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``
        log_info (bool, optional): If ``True``, log some information
            about the weights that are initialized. Default: ``False``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.nn.init import kaiming_normal
        >>> from torch import nn
        >>> net = nn.Sequential(nn.Linear(4, 6), nn.ReLU(), nn.BatchNorm1d(6), nn.Linear(6, 1))
        >>> kaiming_normal(net)
    """
    for name, params in module.named_parameters():
        if params.ndim > 1 and (not learnable_only or learnable_only and params.requires_grad):
            if log_info:
                logger.info(
                    f"Initializing '{name}' with Kaiming Normal (mode={mode}, "
                    f"nonlinearity={nonlinearity}, neg_slope={neg_slope}) | shape={params.shape}"
                )
            nn.init.kaiming_normal_(params.data, a=neg_slope, mode=mode, nonlinearity=nonlinearity)


def kaiming_uniform(
    module: nn.Module,
    neg_slope: float = 0.0,
    mode: str = "fan_in",
    nonlinearity: str = "leaky_relu",
    learnable_only: bool = True,
    log_info: bool = False,
) -> None:
    r"""Initializes the module parameters with the Kaiming uniform
    strategy.

    Args:
    ----
        module (``torch.nn.Module``): Specifies the module to
            initialize.
        neg_slope (float, optional): Specifies the negative slope of
            the rectifier used after this layer (only used with
            ``'leaky_relu'``). Default: ``0.0``
        mode (str, optional): either ``'fan_in'`` or ``'fan_out'``.
            Choosing ``'fan_in'`` preserves the magnitude of the
            variance of the weights in the forward pass. Choosing
            ``'fan_out'`` preserves the magnitudes in the backwards
            pass. Default: ``'fan_in'``
        nonlinearity (str, optional): the non-linear function
            (`nn.functional` name), recommended to use only with
            ``'relu'`` or ``'leaky_relu'``. Default: ``'leaky_relu'``
        learnable_only (bool, optional): If ``True``, only the
            learnable parameters are initialized, otherwise all the
            parameters are initialized. Default: ``True``
        log_info (bool, optional): If ``True``, log some information
            about the weights that are initialized. Default: ``False``

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.nn.init import kaiming_uniform
        >>> from torch import nn
        >>> net = nn.Sequential(nn.Linear(4, 6), nn.ReLU(), nn.BatchNorm1d(6), nn.Linear(6, 1))
        >>> kaiming_uniform(net)
    """
    for name, params in module.named_parameters():
        if params.ndim > 1 and (not learnable_only or learnable_only and params.requires_grad):
            if log_info:
                logger.info(
                    f"Initializing '{name}' with Kaiming uniform (mode={mode}, "
                    f"nonlinearity={nonlinearity}, neg_slope={neg_slope}) | shape={params.shape}"
                )
            nn.init.kaiming_uniform_(params.data, a=neg_slope, mode=mode, nonlinearity=nonlinearity)



r"""This module implements PyTorch sampler functions for data
loaders."""

from __future__ import annotations

__all__ = ["ReproducibleBatchSampler", "PartialSequentialSampler", "PartialRandomSampler"]

from collections.abc import Generator, Iterator, Sized

import torch
from torch.utils.data import BatchSampler, Sampler


class ReproducibleBatchSampler(BatchSampler):
    r"""Implements a reproducible batch sampler.

    This class is inspired from PyTorch Ignite. Internally, this class
    iterates and stores indices of the input batch sampler.

    Args:
    ----
        batch_sampler (``torch.utils.data.BatchSampler``): batch
            sampler same as used with ``torch.utils.data.DataLoader``
        start_iteration (int, optional): Specifies the starting
            iteration. Default: 0.

    Example usage:

    .. code-block:: pycon

        >>> from gravitorch.data.dataloaders.samplers import ReproducibleBatchSampler
        >>> from torch.utils.data import BatchSampler, SequentialSampler
        >>> batch_sampler = BatchSampler(
        ...     SequentialSampler(range(10)),
        ...     batch_size=3,
        ...     drop_last=False,
        ... )
        >>> reproducible_batch_sampler = ReproducibleBatchSampler(batch_sampler)
        >>> list(reproducible_batch_sampler)
        [[0, 1, 2], [3, 4, 5], [6, 7, 8], [9]]
        >>> batch_sampler = ReproducibleBatchSampler(batch_sampler, start_iteration=1)
        >>> list(reproducible_batch_sampler)
        [[3, 4, 5], [6, 7, 8], [9]]
        >>> batch_sampler = ReproducibleBatchSampler(batch_sampler, start_iteration=2)
        >>> list(reproducible_batch_sampler)
        [[6, 7, 8], [9]]
    """

    def __init__(self, batch_sampler: BatchSampler, start_iteration: int = 0) -> None:
        if not isinstance(batch_sampler, BatchSampler):
            raise TypeError(
                "Argument batch_sampler should be torch.utils.data.sampler.BatchSampler"
            )
        if start_iteration < 0:
            raise ValueError("Argument start_iteration should be positive integer")

        self.batch_sampler = batch_sampler
        self._start_iteration = start_iteration

    def __iter__(self) -> Generator:
        yield from list(self.batch_sampler)[self._start_iteration :]

    def __len__(self) -> int:
        return len(self.batch_sampler)


class PartialSequentialSampler(Sampler):
    r"""Implements a partial sequential sampler that samples only the
    first items in the dataset.

    Args:
    ----
        datasource (``Sized``): Specifies the dataset to sample
            from.
        num_samples (int): Specifies the number of samples to draw.
            If the number of samples is bigger than the number of
            samples in the dataset, the number of samples to draw
            is the dataset size.
    """

    def __init__(self, datasource: Sized, num_samples: int) -> None:
        super().__init__(datasource)
        self.datasource = datasource
        if not isinstance(num_samples, int) or num_samples <= 0:
            raise ValueError(
                f"num_samples should be a positive integer value, but got {num_samples}"
            )
        self.num_samples = num_samples

    def __iter__(self) -> Iterator:
        return iter(range(len(self)))

    def __len__(self) -> int:
        return min(len(self.datasource), self.num_samples)


class PartialRandomSampler(PartialSequentialSampler):
    r"""Implements a partial random sampler that samples randomly some
    items of the dataset."""

    def __iter__(self) -> Iterator:
        return iter(torch.randperm(len(self.datasource))[: self.num_samples].tolist())



r"""This module defines the base class for the data loader collators."""

from __future__ import annotations

__all__ = ["BaseCollator", "setup_collator"]

from abc import ABC, abstractmethod
from collections.abc import Callable
from typing import Generic, TypeVar

from objectory import AbstractFactory
from torch.utils.data.dataloader import default_collate

R = TypeVar("R")
T = TypeVar("T")


class BaseCollator(Generic[T, R], Callable[[list[T]], R], ABC, metaclass=AbstractFactory):
    r"""Defines the base class to create a batch of examples."""

    @abstractmethod
    def __call__(self, data: list[T]) -> R:
        r"""Creates a batch given a list of examples.

        Args:
        ----
            data (list): Specifies a list of examples.

        Returns:
        -------
             A batch of examples.
        """


def setup_collator(collator: Callable[[list[T]], R] | dict | None) -> Callable[[list[T]], R]:
    r"""Sets up a data loader collator.

    Args:
    ----
        collator (``Callable`` or dict or None): Specifies the
            data loader collator or its configuration. If ``None``,
            the default data loader collator is used.

    Returns:
    -------
        ``Callable``: The data loader collator.
    """
    if collator is None:
        collator = default_collate
    if isinstance(collator, dict):
        collator = BaseCollator.factory(**collator)
    return collator
